/*******************************************************************************
 server_frames.js: server file to handle frame enabling, associating etc. (need restructuring)

 $Author: dmc $
 $Revision: 2898 $
 $Date: 2021-03-19 16:08:01 -0300 (Fri, 19 Mar 2021) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/server_frames.js $
*******************************************************************************/

// handle enabling and disabling frames
// functions to handle association between targets and frames

// beams.beamlisteners is hashed by beamip; points to a TargetListener

///

var logShotToFile = function( shot, name ) {
	if ( !shot.valid ) {
		return;
	}

	var msg = shot.ts + "," + shot.res.x + "," + shot.res.y + "," + shot.res.v;
	msg += "," + shot.res.theta + "," + shot.res.phi + "," + shot.res.temp + "," + shot.mcucount;
	for ( var i = 0; i < shot.mics.length; i++ ) {
		msg += "," + shot.mics[ i ];
	}
	var logger = log4js.getLogger( name );
	logger.info( msg );
};

///

beams = require("./smt.common.lib/beam.js");   // returns EventEmitter
beamslistener = new beams.BeamsListener();
var mic_mapping = require('./smt.common.lib/mic_mapping.js');

var log4js = require( "log4js" );
var log4 = log4js.getLogger( "server-frames.js" );

var fullborejs = require("./smt.common.lib/fullbore.js");
var calc_shot = require("./smt.common.lib/calc_shot.js");
var calc_shot_nr = require("./smt.common.lib/calc_shot_nr.js");
var calc_shot_mn = require("./smt.common.lib/calc_shot_mn.js");
//eval(fs.readFileSync("./smt.common.lib/common.server.js")+""); // common utils
var common_serverjs = require('./smt.common.lib/common.server.js');

function frames_init() {
	serialport_init(); // if we are a G2, connect to the MCU
	faListenerInit(); // if we are a SOLO, create listener

	// just make sure targets are not green
	// association and devices are retained - make sure the devices are kept in the data.devices array!
	for (var i in data.frames) {
		var frame = data.frames[i];		
		frame.connected = false;
	}

}

// permanently delete a frame
function frame_delete(frame_id) {
	log4.debug("frame_delete", frame_id);

	// cannot delete the last frame
	if (numberof(data.frames) == 1) return;
	
	// remove association to device
	frame_unassociate(frame_id);
	
	// delete archived strings that used this frame
	for (var i in archive) {
		var str = archive[i];
		if (str.frame_id == frame_id) {
			delete archive[i];
		}
	}

	// remove from data
	var frame = data.frames[frame_id];
	delete data.strings[frame.string_id];
	delete data.frames[frame_id];
	
	// force full update to clients
	broadcast_data();
	
}

function frame_create(frame_id) {
	log4.debug( "frame_create" );
	
	var frame = {}; // create new frame as copy of current frame
	update_object(frame, data.frames[frame_id]);
	frame._id = get_unique_id();
	frame.name = "new";
	frame.scale = false;
	frame.geom.x_cal = 0;
	frame.geom.y_cal = 0;
	frame.assoc_device_id = 0;
	frame.assoc_device_id_last = 0;
	delete frame.pickmepickme; // there can only be one special SOLO-autoselected frame, and you are not it!

	var str = {}; // create new string with minimum content
	str._id = get_unique_id();
	str.name = "";
	str.face_id = data.strings[frame.string_id].face_id; // use the same face id
	str.frame_id = frame._id;
	str.shots = [];
	str.ts = Date.now();
	
	frame.string_id = str._id;

	// apply string_params from frame
	if (frame.string_params) update_object(str, frame.string_params);
	
	data.frames[frame._id] = frame;
	data.strings[str._id] = str;
	
	// force full update to clients
	broadcast_data();

}

function frame_create_SOLO() {
	//'use strict';
	log4.debug( "frame_create" );
	
	var frame = {}; // create new frame as copy of current frame
	update_object(frame, {
			"_id": "SOLO",
			"name": "SOLO",
			"geom": {
				"mic_width": 1700,
				"mic_height": 1700,
				"chrono_spacing": 170,
				"x_cal": 0,
				"y_cal": 0,
				"temp_cal": 3,
				"sens_geom": "STD",
				"wiring_harness": "SOLO",
				"mic_mapping": "58472613"
			},
			"enabled": true,
			// "string_id": 1516033533383,
			"connected": true,
			"assoc_device_id": 0,
			"assoc_device_id_last": 0,
			"connection_type": "TCP",
			"string_params": {
				"sighter_mode": "freeform",
				"shots_mode": -1,
				"shooter_count": 1
			}
		}
	);
	// frame._id = get_unique_id();
	// frame.name = "new";
	// frame.scale = false;
	// frame.geom.x_cal = 0;
	// frame.geom.y_cal = 0;
	// frame.assoc_device_id = 0;
	// frame.assoc_device_id_last = 0;

	var str = {}; // create new string with minimum content
	str._id = get_unique_id();
	str.name = "";
	str.face_id = Object.keys(data.faces)[0]; // use first targetface
	str.frame_id = frame._id;
	str.shots = [];
	str.ts = Date.now();
	
	frame.string_id = str._id;

	// apply string_params from frame
	if (frame.string_params) update_object(str, frame.string_params);
	
	data.frames[frame._id] = frame;
	data.strings[str._id] = str;
	
	// frame_associate_G2("SOLO","1234");
	
	// force full update to clients
//	broadcast_data();
}

function frame_set_enabled(frame_id, enabled) {
	var frame = data.frames[frame_id];
	if (frame.enabled == enabled) return;
	
	var crud = {};
	crud.frames = {};
	crud.frames[frame_id] = { enabled: enabled };
	apply_crud(crud);
	
	// enabling: check to see if last device is available
	if (enabled && frame.assoc_device_id_last) {
		var device = data.devices[frame.assoc_device_id_last];
		if (device && !device.assoc_frame_id) {
			frame_associate(frame_id, frame.assoc_device_id_last);
		}
	}
	
	// disabling: remove association
	if (!enabled && frame.assoc_device_id) {

		var device_id = frame.assoc_device_id;
		
		frame_unassociate(frame_id);
		
	}

}

function frame_associate(frame_id, device_id) {
	frame_associate_specific(frame_id, device_id, "G1");
}

function frame_associate_G2(frame_id, device_id) {
	frame_associate_specific(frame_id, device_id, "G2");
}

function frame_associate_specific(frame_id, device_id, device_type) {
	log4.debug("frame_associate", frame_id, device_id, device_type);

	// get objects
	var device = data.devices[device_id];
	
	// if device does not exist, create and start a listener (for G1)
	if (!device) {
		
		// create this device
		var crud = {};
		crud.devices = {};
		crud.devices[ device_id ] = { type: device_type, allow_association: true };

		if ( device_type !== "SOLO" ) {
			crud.devices[ device_id ].ip = device_id;
		} else {
			crud.devices[ device_id ].type = "SOLO";
		}

		apply_crud( crud );
		
		device = data.devices[ device_id ];
				
//console.log("\n\n frame_associate_specific create device_id=",device_id,"device=",device);

		frame_create_listener( device );
	}
	
	var frame = data.frames[frame_id];
	
	// remove any current associations
	if (frame.assoc_device_id) frame_unassociate(frame_id);
//	if (device.assoc_frame_id) frame_unassociate(device.assoc_frame_id);

	log4.debug( "associating " + frame.name + " to " + device.internalId );

	// use crud so it is sent to all clients
	var crud = {};
	crud.frames = {};
	crud.frames[frame_id] = {
		assoc_device_id: device_id,
		assoc_device_id_last: device_id // for auto association later
	};
	var assoc_frame_ids_list = device.assoc_frame_ids_list || {};
	assoc_frame_ids_list[frame_id] = {};
	crud.devices = {};
	crud.devices[device_id] = {
		assoc_frame_id: frame_id, assoc_frame_ids_list: assoc_frame_ids_list
	};
	
	apply_crud(crud);
	
}

function frame_unassociate(frame_id) {
	log4.debug("frame_unassociate", frame_id);

	// get objects
	var frame = data.frames[frame_id];
	if (!frame.assoc_device_id) return;
	var device = data.devices[frame.assoc_device_id];

	if (!device) return; // should not be needed

	log4.debug( "unassociating " + frame.name + " from " + device.internalId );

	// use crud so it is sent to all clients
	var crud = {};
	var assoc_frame_ids_list = device.assoc_frame_ids_list || {};
	//console.log("DMC: assoc_frame_ids_list",assoc_frame_ids_list,frame_id);
	if (assoc_frame_ids_list[frame_id])delete assoc_frame_ids_list[frame_id];
	var pick_frame_id = Object.keys(assoc_frame_ids_list)[0];
	if ('undefined' == typeof pick_frame_id) pick_frame_id = 0;
	crud.devices = {};
	crud.devices[frame.assoc_device_id] = {
		assoc_frame_id: pick_frame_id, assoc_frame_ids_list: assoc_frame_ids_list
	};
	crud.frames = {};
	crud.frames[frame_id] = {
		assoc_device_id: 0
	};
	
	apply_crud(crud);

}

function frame_simulate_shot(frame_id, x, y, v, theta, phi, temp) {
	log4.debug("frame_simulate_shot", frame_id, x, y, v, theta, phi, temp);
	
	// send to the frame's associated target
	var frame = data.frames[frame_id];
	if (frame.assoc_device_id) {
		var device = data.devices[frame.assoc_device_id];
		var req = {
		  body: {
		    frame_name: frame.name,
		    x: x, y: y, v: v, theta: theta, phi: phi, temp: temp,
		  }
		};
		var calcd_shot_timings = namedframe_calc_timings_sim_shot(req);
		if (calcd_shot_timings) {
			frameid_emit_shot_direct(frame_id,calcd_shot_timings);
		}
		log4.debug("sent shot to frame");
	} else {
		log4.debug("this frame is not associated with a target!");
	}
}

function find_frame_id(frame_name) {
	var frames=data.frames;
	for (frame_id in frames) {
		if (frame_name == frames[frame_id].name) return frame_id;
	}
}

function namedframe_emit_shot_req(req) {
	var frame_name = req.body.frame_name;
	// find frame by name
	var frame_id = find_frame_id(frame_name);
	// send to the frame's associated target
	frameid_emit_shot_direct(frame_id,req.body);
}

function frameid_emit_shot_direct(frame_id,shotdata) {
	// shotdata must contain:
	/*
		{
			"ts": 1499111351103,
			"testshot_id":"16:49:11 T3 Shit 12+ shots #--/42, T3 Lodi shot#5",
			"source": "G2",
??			"frame_name":"T1-remapped",
			"shot": {
				"shotnum": 12,
?? check for multiple application of tempcal
				"G2_temp_rdg": 25.7,
				"counters": [3949, 188, 9798, 3342, 0, 5926, 9106, 5683  ]
			}

		}

	*/
	
	var frame = frame_id && data.frames[frame_id];
//console.log("DMC: measToFrame shotdata.shot=",shotdata.shot);
	measToFrame(shotdata, frame);
	log4.debug("frameid_emit_shot_direct sent shot to measToFrame",shotdata);
}


var calc_shot_timings_num=1;

//function namedframe_calc_timings_sim_shot(frame_name, x, y, v, theta, phi, temp) {
function namedframe_calc_timings_sim_shot(req) {
	var
	  frame_name = req.body.frame_name,
	  x = req.body.x, y = req.body.y, v = req.body.v,
	  theta = req.body.theta, phi = req.body.phi,
	  temp = req.body.temp
	;

	var t1=accuratetime();

	log4.debug("namedframe_calc_timings_sim_shot", frame_name, x, y, v, theta, phi, temp);
	// find frame by name
	var frame_id = find_frame_id(frame_name);
	// send to the frame's associated target
	var frame = frame_id && data.frames[frame_id];
	if (!frame) return;
	var logshot = {
		ts: Date.now(),
		frame_name: frame_name,
		source: 'sim_sw',  // TODO: sim_sw_3
		sim_data: {x:x, y:y, v:v, theta:theta, phi:phi, temp:temp, geom: frame.geom},
		shot: { shotnum:calc_shot_timings_num++, },
		mic_mapping: mic_mapping.get_mic_mapping_frame(frame),
		geom: frame.geom
	}

	calc_shot.simulate_timings(logshot); // new method
	delete logshot.geom; // only needed temporarily for calc'n
	log4.debug("namedframe_calc_timings_sim_shot calcd result",logshot);
	//logshot.e_t_ms = 1000*(accuratetime() - t1);
	return logshot;
}

function frame_get_listener( device ) {
	return beamslistener.get_listener( device );
}

function frame_close_listener( device ) {
	var b_listener = beamslistener.get_listener( device );

	if ( b_listener !== undefined ) {
		b_listener.stop_poller();
		beamslistener.delete_listener( device );
	}
	
}

// create a listener for a target-type device
// MAY NOT be assigned to a frame
function frame_create_listener( device ) {
	
	if ( device.type === "SOLO" ) {
		log4.debug( "ignore frame_create_listener() for SOLO" );
		return;
	}

	log4.debug( "creating listener for " + device.internalId );
//console.log("\n\nframe_create_listener",device.internalId);

	// create the listener
	//var b_listener = beamslistener.get_listener(device.ip, (device.type == "G2" || device.smtdevtype == "G2") ? "TCP" : "");
	var b_listener = beamslistener.get_listener( device );
//console.log("\n\nframe_create_listener --->",b_listener,"\n\n");

	b_listener.beamevents.on("tgt_heartbeat",
		function() {
			
			var conn = ( b_listener.tgt_stat == "tgt_there" );
			var frame = null;
			if ( device.assoc_frame_id ) frame = data.frames[ device.assoc_frame_id ];
			
			if ( frame && frame.connected != conn ) {
				crud = {};
				crud.frames = {};
				crud.frames[ frame._id ] = { connected: conn };
//console.log("beamevents.on  tgt_heartbeat connected:",conn);
				apply_crud( crud );
			}

		}
	);

	// receive shot data from target, compute position/velocity, send to recv_shot()
	b_listener.beamevents.on("meas-from-beam", function(beam_measurement) {
		measFromBeam(beam_measurement, device); // G1 and G2 measurements go through here
	});

	
	b_listener.start_poller();

	return b_listener;

}

// we want to aggressively trim the precision of the numbers delivered
// in a shot string.
function trim_num(nn,ndigs) {
		if (typeof ndigs == 'undefined' || ndigs < 0 || ndigs > 5) ndigs = 0;
		if (typeof nn == 'number') return +nn.toFixed(ndigs); // '+' needed to return 'number' not 'string'
		return nn;
}

global.qndEventEmitter.on( "measFromBeam", function( data ) {
	measFromBeam( data.beamMeasurement, data.srcDevice ); // SOLO measurements go through here
} );

function measFromBeam(beam_measurement, srcDevice) { // both SOLO and G2 measurements call us
	if (!srcDevice || !srcDevice.assoc_frame_id) return;
	// backward compatibility: might have only assoc_frame_id defined
	var frame1 = data.frames[srcDevice.assoc_frame_id];
	measToFrame(beam_measurement, frame1);
	for (var i in srcDevice.assoc_frame_ids_list) {
		var frame = data.frames[i];
		if (frame1 == frame) continue; // already been sent out, don't send twice!
		measToFrame(beam_measurement, frame);
	}
}
var previous_shot_start = undefined;

function measToFrame(beam_measurement, frame) {
	if (previous_shot_start) {
		// console.log('measToFrame - previous shot ' + _performance(previous_shot_start) + 'ms)');
	}
	var start = _performance();
	previous_shot_start = start; // for our next time through

	if (!frame) return;
	var str = data.strings[frame.string_id];
	var face = data.faces[str.face_id];

	log4.debug("measToFrame() shot on frame ", frame.name);

	var shot = beam_measurement;
	shot.geom = frame.geom; // TODO: don't create shot.geom
	var sens_geom = shot.geom.sens_geom;
	var geom_is_chamber = "CHAMBER" == sens_geom;
	var geom_is_hbar = 0 == sens_geom.indexOf("HBAR");
	var geom_is_std = 0 == sens_geom.indexOf("STD"); // STD*
	var geom_is_stdNR = "STDNR" == sens_geom; // do closed-form iterative search, then NR
	var geom_is_stdMN = "STDMN" == sens_geom; // do closed-form iterative search, then MN
	var geom_is_IMPACT = "IMPACT" == sens_geom;

	// DMC bug, I think - correct source of mic_mapping should be frame, not shot (matters mostly for replay)
	//shot.mic_mapping = mic_mapping.get_mic_mapping(shot); // needed in calc_shot() recv_shot() etc, so that mic_mapping() calls work
	shot.mic_mapping = mic_mapping.get_mic_mapping_frame(frame); // needed in calc_shot() recv_shot() etc, so that mic_mapping() calls work
	shot.res = {};
	log4.debug(shot);
		// device_type, clock, temp, soundspeed
		calc_shot.calc_initial(shot.res,shot); // needed by calc_shot_nr.* too
		// console.log('measToFrame - calc_initial (' + _performance(start) + 'ms)');

		if (geom_is_std) {
			// 15ms off target, 100ms on target
			var t1=accuratetime();
			calc_shot.solve_iterative(shot);
			shot.res.method = "ITER";
			shot.res.e_t_ms = Math.round(1000*(accuratetime() - t1));

			if (geom_is_stdNR && shot.valid) {
				// Newton-Raphson search method
				var t1=accuratetime();
delete shot.whack_raw_v;
delete shot.whack_v;
delete shot.whack_xy_err;
delete shot.whack_v_err;
delete shot.x;
delete shot.y;
//delete shot.v;
				calc_shot_nr.solve_nr_8mic_cube(shot);
				shot.res_nr.e_t_ms = Math.round(1000*(accuratetime() - t1));
shot.res=shot.res_nr;  delete shot.res_nr;
			}

//			if (1 && Math.abs(shot.res.x) < 1000 && Math.abs(shot.res.y) < 1000 ) { // MN test
			if (geom_is_stdMN && shot.valid) { // MN test
				// Minimization search method
				var t1=accuratetime();
delete shot.whack_raw_v;
delete shot.whack_v;
delete shot.whack_xy_err;
delete shot.whack_v_err;
delete shot.x;
delete shot.y;
//delete shot.v;
				calc_shot_mn.solve_mn(shot);
				shot.res_mn.e_t_ms = Math.round(1000*(accuratetime() - t1));
shot.res=shot.res_mn; // delete shot.res_nr;
shot.res.method="MN";

			}



		} else if (geom_is_hbar) {
			var t1=accuratetime();

			calc_shot_nr.solve_nr_hbar(shot);
			shot.res_hbar.e_t_ms = Math.round(1000*(accuratetime() - t1));
shot.res = shot.res_hbar; shot.res_hbar = "moved to res";
		} else if (geom_is_IMPACT) {
			var t1=accuratetime();

			calc_shot.solve_impact(shot);
			shot.res.method = "IMPACT";
			shot.res.e_t_ms = Math.round(1000*(accuratetime() - t1));

		}
		
		// console.log('measToFrame - calc done (' + _performance(start) + 'ms)');
		// extract trimmed data from shot
		if (shot.res.xy_valid) {
			shot.x = trim_num(shot.res.x,2); // may not exist
			shot.y = trim_num(shot.res.y,2);
		} else {
			delete shot.x; delete shot.y; // for invalid (x,y), we make *SURE* that shot.x and shot.y do not exist!!
		}
		shot.v = trim_num(shot.res.v, 1); // always exists, 0 when invalid
		shot.tempC = trim_num(shot.res.temp, 1);
		if (shot.res.temp_invalid) shot.temp_invalid = true; // true when fallback temp used
		shot.num = shot.shot.shotnum;
		shot.GPIO18 = shot.shot.GPIO18;
		shot.mics = shot.res.mic;
		
		if (shot.res.v) shot.res.v = trim_num(shot.res.v, 2);
		if (shot.res.x) shot.res.x = trim_num(shot.res.x, 2);
		if (shot.res.y) shot.res.y = trim_num(shot.res.y, 2);
		if (shot.res.v13) shot.res.v13 = trim_num(shot.res.v13, 2);
		if (shot.res.v26) shot.res.v26 = trim_num(shot.res.v26, 2);
		if (shot.res.v47) shot.res.v47 = trim_num(shot.res.v47, 2);
		if (shot.res.v58) shot.res.v58 = trim_num(shot.res.v58, 2);
		if (shot.res.raw_v13) shot.res.raw_v13 = trim_num(shot.res.raw_v13, 2);
		if (shot.res.raw_v26) shot.res.raw_v26 = trim_num(shot.res.raw_v26, 2);
		if (shot.res.raw_v47) shot.res.raw_v47 = trim_num(shot.res.raw_v47, 2);
		if (shot.res.raw_v58) shot.res.raw_v58 = trim_num(shot.res.raw_v58, 2);
		if (shot.res.soundspeed) shot.res.soundspeed = trim_num(shot.res.soundspeed, 2);
		if (shot.res.theta) shot.res.theta = trim_num(shot.res.theta, 4);
		if (shot.res.phi) shot.res.phi = trim_num(shot.res.phi, 4);
		if (shot.res.xy_error) shot.res.xy_error = trim_num(shot.res.xy_error, 4);

		if (shot.shot.mcucount) shot.mcucount = shot.shot.mcucount;

		// delete unneeded memory
		var debugging = false;
		if (data.config.option_debuginfo) debugging=true; // GUI checkbox
		if (!debugging) {
			delete shot.res.mic; // keep res for debugging for now
			//delete shot.res.raw_counters;
			delete shot.geom;
//			delete shot.sim_data;

//			delete shot.shot; // needed for simulating a shot
		}
		
		//if (data.config.option_trimres) delete shot.res; // aggressive trimming
//log4.debug("DMC meas-from-beam, shot after cleanup",shot);
//process.exit(1);

		// handle shot on server
		// console.log('measToFrame - recv_shot begin (' + _performance(start) + 'ms)');
		common_serverjs.recv_shot(shot, frame._id, data);
		// console.log('measToFrame - recv_shot end (' + _performance(start) + 'ms)');

	logShotToFile( shot, frame.name );
	// console.log('measToFrame - logShotToFile done  (' + _performance(start) + 'ms)',shot.valid);

}

function accuratetime() {
	var res = process.hrtime();
	log4.debug(res);
	return res[0] + res[1] * 1e-9;
}

function val_num(n) {
	if (typeof n == 'undefined' ||
	(typeof n != 'number') ||
	(typeof n == 'number' && (isNaN(n) || !isFinite(n)))
	)
	  return false;
	return true;
}

