/*******************************************************************************
 $Author: dmc $
 $Revision: 3135 $
 $Date: 2022-10-20 00:17:17 -0300 (Thu, 20 Oct 2022) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/viewer_users.js $
*******************************************************************************/
// user-related client side helper functions

var cur_user_id = null;
function cur_user() { // current user logged in
	if (cur_user_id == null) return null; // guest
	return g_users[cur_user_id]; // the user
}

function user_select(id) { // switch to this user
	log("switching to user " + id);
	cur_user_id = id; // need to update after this
	if (id) set_cookie("recentuser", id); // remember on login
	else clear_cookie("recentuser"); // clear on logout
}

function user_display_name(id) {
	return g_users[id] && ((g_users[id].data.realname || "") + ("[" + g_users[id].data.name + "]")); // always show userid; show real name if possible
}

function admin_check() {
	return cur_user() && cur_user().data.name == "admin";
}

function ro_check() {
	return cur_user() && cur_user().data.is_range_officer;
}

function extended_check() {
	return cur_user() && cur_user().data.is_extended_user;
}

function ordinary_user_check() {
	return !admin_check() && !ro_check() && !extended_check();

}

// whether save string button should appear
function user_can_save_string() {
	if (!g_config.option_access_mode) return true; // freeform
	if (admin_check() || ro_check()) return true; // admin and ro can always
	if (!cur_user()) return false; // must be logged in
	if (user_is_scorer() && g_config.option_access_mode < 3) return true; // scorekeeper
	if (g_config.option_access_mode == 1 && user_is_shooter() >= 0) return true; // shooter
	return false;
}

// whether shot bubble is accessible and can manage the shots
function user_can_affect_shots() {
	return user_can_save_string();
}

// whether or not to display invalid shots for debugging/calibrating/etc
// show in frame panel or in review mode with option set
// if admin only, requires admin_check()
function display_admin_mode(admin_only) {
	if (admin_only && !admin_check()) return false;
	return (toolbar.active_panel == toolbar_system)
		|| (!live_mode() && toolbar_review.display_admin)
//		|| toolbar.active_panel == toolbar_info // DMC 13-July-2017
	;
}

// whether or not to display invalid shots (crossfires, off-target etc)
// used in "i"-button display mode
function display_info_mode() {
	return (!live_mode() && toolbar_review.display_admin) || toolbar.active_panel == toolbar_info;
}

// if user is shooter, return which one (0, 1, 2)
// otherwise return -1
// leave empty parameters for current string / user
// even_if_disabled: check disabled shooter slots
function user_is_shooter(str_id, user_id, even_if_disabled) {
	if (!even_if_disabled && !g_config.use_role_shooter) return -1;
	if (!str_id && cur_string()) str_id = cur_string().id();
	if (!user_id) {
		if (!cur_user()) return -1;
		user_id = cur_user().id();
	}
	var str = g_strings[str_id];
	if (!str) return -1;
	var count = str.data.shooter_count || 1;
	if (even_if_disabled) count = 3;
	var shooters = str.data.role_shooter;
	if (!shooters) return -1;
	for (var i = 0; i < count; i++) {
		if (shooters[i] == user_id) return i;
	}
	return -1;
}

// whether user is the assigned scorekeeper
// leave empty parameters for current string / user
// even_if_disabled: check disabled shooter slots
function user_is_scorer(str_id, user_id) {
	if (!g_config.use_role_scorer) return 0;
	if (!str_id && cur_string()) str_id = cur_string().id();
	if (!user_id) {
		if (!cur_user()) return 0;
		user_id = cur_user().id();
	}
	var str = g_strings[str_id];
	if (!str) return 0;
	return str.data.role_scorer == user_id;
}

// must specify string and user, 0 to unassign
function user_assign_shooter(str_id, user_id, index) {

	var string = g_strings[str_id];
	if (!string) return;

	var role_shooter = [];
	var fclass_target = [];
	
	var user = g_users[user_id];
	
	if (!user) { // unassign

		role_shooter[index] = 0;
		
	} else { // assign
	
		role_shooter[index] = user_id;
		fclass_target[index] = user.data.fclass_target ? 1 : 0; // copy from user profile
	
		// this user is already assigned as another shooter: unassign
		var shooter_index = user_is_shooter(str_id, user_id, true);
		if (shooter_index >= 0) {
			role_shooter[shooter_index] = 0;
		}
		
		if (user_is_scorer(str_id, user_id)) {
			string.crud({ role_scorer: 0 });
		}
	}

	string.crud({role_shooter: role_shooter, fclass_target: fclass_target});
	
	g_socket.emit("recalc_shot_scores", str_id);
	
}

// must specify string and user, 0 to unassign
function user_assign_scorer(str_id, user_id) {

	var string = g_strings[str_id];
	if (!string) return;

	var user = g_users[user_id];
	
	if (!user) { // unassign

		string.crud({ role_scorer: 0 });
		return;
		
	} else { // assign
	
		string.crud({ role_scorer: user_id });
	
		// this user is already assigned as another shooter: unassign
		var shooter_index = user_is_shooter(str_id, user_id, true);
		if (shooter_index >= 0) {
			var role_shooter = [];
			role_shooter[shooter_index] = 0;
			string.crud({role_shooter: role_shooter});
		}
		
	}
	
}


// only for live mode
function get_target_ready_status(str_id) {
	if (!live_mode()) return {};
	var str = g_strings[str_id];
	if (!str) return {};
	
	var sdata = str.data;
	
	if (!str.frame().data.connected) {
		return {status: 0, text: "Target disconnected!", css: "text_red"};
	}
	
	var shooter_count = sdata.shooter_count || 1;
	
	// check scorer
	if (g_config.use_role_scorer && !sdata.role_scorer) {
		return {status: 0, text: "Scorekeeper not assigned!", css: "text_red"};
	}

	// check shooters
	if (g_config.use_role_shooter) {
		
		var has = 0;
		
		if (sdata.role_shooter) {
			for (i = 0; i < sdata.role_shooter.length && i < shooter_count; i++) {
				if (sdata.role_shooter[i]) has++;
			}
		}
		
		if (has != shooter_count) {
			return {status: 0, text: "Shooters not assigned!", css: "text_red"};
		}
		
	}

	// check final score
	if (g_config.option_total_score && g_config.shots_mode > 0) {
		var has_final = 0;
		var has_toomany = 0;
		for (var i = 0; i < shooter_count; i++) {
			var score = calc_string_score(sdata, i);
			if (score.count == g_config.shots_mode) has_final++;
			else if (score.count > g_config.shots_mode) has_toomany++;
		}
		
		if (has_toomany) return {status: 0, text: "Too many shots!", css: "text_red"};
		if (has_final == shooter_count) return {status: 0, text: "All scores final!", css: "text_green"};
	}

	return {status: 1, text: "Ready to fire!", css: "text_green"};
}

function cur_event() { // return g_events[id] for active event, or null
	var use_event = g_config.event_enabled && g_config.event_data.event_id >= 0 && g_config.event_data.match_id >= 0 && g_config.event_data.stage_id >= 0;
	if (!use_event) return null;
	return g_events[g_config.event_data.event_id];
}

// return: 0 if event has registered shooters but this user is not
// -1 if the event registration is not active
// for registered shooters, the object
function user_is_registered(user_id) {
	if (!user_id && cur_user()) user_id = cur_user().id();
	var event = cur_event();
	if (!event || !numberof(event.shooters)) return -1;
	if (!event.shooters[user_id]) return 0;
	return event.shooters[user_id];
}


// whether main target (cur_string) should be f-class or not
function draw_fclass_target(string_id) {
	
	// target face has no f-class zones defined
	if (!cur_string().face().data.scoringzones_f) return false;
	
	// string has option set
	if (cur_string().data.fclass_target && cur_string().data.fclass_target[cur_shooter]) {
		return true;
	}
	
	// default
	return false;
}























