/*******************************************************************************
 server_storage.js: textfile-based storage of our data.

 $Author: dmc $
 $Revision: 2633 $
 $Date: 2019-07-24 23:28:23 -0300 (Wed, 24 Jul 2019) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/server_storage.js $
*******************************************************************************/

var path = require("path");
var zlibjs = require("zlibjs");
var disk = undefined;
try {
  disk = require("storage-device-info");
} catch (e) {
  log("cannot load storage-device-info; disk space stats not available possible", e);
}


var _storageWriteData = false;
var _storageWriteArchive = false;
var _archive = {};
var _storageWriteDataRunning = false;
var _storageWriteArchiveRunning = false;
var _data = {};
var _performanceLogEnabled = true;
var _storageWriteDataCallbacks = [];
var _storageWriteArchiveCallbacks = [];

var _performance = require('./smt.common.lib/lib/utils/performance.js');

function storage_update_diskspace() {
	disk && disk.getPartitionSpace(".", function(error, space) {
		if (error) {
			log("storage_update_diskspace() error, err=",error);
		} else {
			var free_shots = Math.floor(space.freeMegaBytes * 1024 * 1024 / 40);
			data.config.diskspace = free_shots; // do not use crud so clients don't refresh so often
			log("storage_update_diskspace(), free space=", space.freeMegaBytes,"MB ==> ~"
			,data.config.diskspace,"shots");
		}
	});
}

// check that data exists, otherwise create template files
function storage_init( cb ) {

	if (!fs.existsSync("data") || !fs.existsSync("data/data.txt")) {
		log("creating new default data files");
		if (!fs.existsSync("data")) fs.mkdirSync("data");
		storage_write_data(JSON.parse(fs.readFileSync("initial-data/data.txt")));
		storage_write_archive({});
		if (global.is_G2) fs.writeFileSync("data/settings.mcu.json", fs.readFileSync("initial-data/settings.mcu.json"));
	}
	
	// fix individual files just in case data exists but not these
	if (!fs.existsSync("data/archive.txt")) storage_write_archive({});
	if (!fs.existsSync("data/settings.mcu.json")) fs.writeFileSync("data/settings.mcu.json", fs.readFileSync("initial-data/settings.mcu.json"));
	if ( typeof cb === "function" ) {
		cb( ); // no error
	}
}

// misc cleanup to save disk space
function storage_cleanup() {
	
	exec("rm -f /tmp/*.tgz"); // any leftover downloaded files
	
	// DONT DELETE ANYTHING FROM DATA FOLDER --- too dangerous
/*
	// delete any strings or images in data that are not referenced
	var dir = fs.readdirSync("data");

	function dir_remove(str) {
		var i = dir.indexOf(str);
		if (i >= 0) dir.splice(i, 1);
	}

	dir_remove("data.txt");
	dir_remove("archive.txt");
	dir_remove("settings.mcu.json");
	
	// keep any active face images
	for (var i in data.faces) {
		if (data.faces[i].image) {
			var image = data.faces[i].image.src;
			var start = image.lastIndexOf("/");
			if (start >= 0) {
				image = image.substring(start + 1);
			}
			dir_remove(image);
		}
	}
	
	// keep anything in archive
	for (var i in archive) {
		var str = "string-" + i;
		dir_remove(str + ".zip");
		dir_remove(str + ".txt");
	}
	
	// remove anything left
	if (dir.length > 0) {
		log("removing unreferenced files from data:");
		log(dir);
		for (var i in dir) {
			fs.unlinkSync("data/" + dir[i]);
		}
	}
*/
}

// problem is, that stringfy blocks the nodejs process...
// possible solution: http://stackoverflow.com/questions/34039072/avoiding-node-js-event-loop-block-while-running-json-stringify

// use this when the intention is for a full save
function storage_write_all(callback) {
    var startPref = performanceLog();
    var called = false;
    var onFinish = function() {
        if (called) {
            performanceLog('storage_write_all() finished...', startPref);
            if (callback && typeof callback == 'function') callback();
        }
        called = true;
    };
    
    storage_write_data(data, onFinish);
    storage_write_archive(archive, onFinish);
}


// write frames, faces, and active strings to file
function storage_write_data(dataContent, callback) {
    _data = dataContent;
    _storageWriteData = true;
    
    if (callback) {
        _storageWriteDataCallbacks.push(callback);
    }
}

// write faces to file
function storage_write_faces(data) {
	try {
		// var start = _performance();
		var str_faces = JSON.stringify(data.faces, null, "\t");
		// console.log('storage_write_data - stringify faces (' + _performance(start) + 'ms)', str_faces.length, 'bytes');
        write_file_and_backup("data","faces.txt", str_faces); // no callback
	} catch (e) {
		log("error writing 'faces' to file", e);
	}
}


// cloned over to new_storage.js.... eventually need to migrate over there
function write_file_and_backup (dirname, filename, data_string, callback) {
	var mainfilename = dirname+"/"+filename;
	var bak_filename = mainfilename + ".bak";
	var bak_delay = 5000; // write backup file after main file has had a chance to be flushed out
    fs.writeFile(mainfilename, data_string, function(error) {
    	setTimeout(function() {
		    fs.writeFile(bak_filename, data_string, function() {
		        console.log("finished writing",bak_filename);
		    }); // no callback
    	}, bak_delay);
    	
    	if (typeof callback == 'function') {
    		callback(error);
    	}
    });
}

function storage_write_data_immediate(dataContent, callback) {
    var start = performanceLog('storage_write_data_immediate start save data...');
	var tmp_faces = dataContent.faces;
	dataContent.faces = undefined;
	var str = JSON.stringify(dataContent, null, "\t"); // excluding faces
	dataContent.faces = tmp_faces; // restore
    performanceLog('data stringified... (length: ' + str.length + ')', start);
    write_file_and_backup("data","data.txt", str, function(error) {
    	if ('function' == typeof callback) {
    		callback(error);
    	}
        performanceLog('storage_write_data_immediate data saved...'+(error?'error='+error:''), start);
    });
}

setInterval(function() {
    if (_storageWriteData && !_storageWriteDataRunning) {
    	try {
            _storageWriteDataRunning = true;
            _storageWriteData = false;
            storage_write_data_immediate(_data, function(error) {
                if (!error) {
                    storage_update_diskspace();
                }
                while(_storageWriteDataCallbacks.length) {
                    var t = _storageWriteDataCallbacks.shift();
                    t(!error, error);
                }
                _storageWriteDataRunning = false;
            });
        } catch (e) {
            log("error writing data to file", e);
            _storageWriteDataRunning = false;
        }
    }
    
    if (_storageWriteArchive && !_storageWriteArchiveRunning) {
        try {
            _storageWriteArchive = false;
            _storageWriteArchiveRunning = true;
            fs.writeFile("data/archive-XXX.txt", '{"no-longer-saving":"this data"}', function(error) {
//            fs.writeFile("data/archive.txt", JSON.stringify(_archive, null, "\t"), function(error) {
                if (!error) {
                    storage_update_diskspace();
                }
                _storageWriteArchiveRunning = false;
                while(_storageWriteArchiveCallbacks.length) {
                    var t = _storageWriteArchiveCallbacks.shift();
                    t(!error, error);
                }
            });
        } catch (e) {
            log("error writing data to file", e);
            _storageWriteArchiveRunning = false;
        }
    }
    
}, 5000);

// read old strings without shots from file
function storage_read_archive() {
	try {
		var archive = JSON.parse(fs.readFileSync("data/archive.txt"));
		
		// delete any invalid strings
		for (var i in archive) {
			var str = archive[i];
			if (!data.frames[str.frame_id] || !data.faces[str.face_id]) {
				log("archived string has invalid frame or face!");
				delete archive[i];
			}
		}
		
		// todo: delete any strings that the file does not exist for
		// todo: add any valid strings that there are files for
		
		return archive;
		
	} catch (e) {
		log("error reading archive from file", e);
	}
	return {}; 	// fall-through; don't crash caller if we can't read archive.txt
}

// write string archive
function storage_write_archive(archiveContent, callback) {
    _archive = archiveContent;
    _storageWriteArchive = true;

    if (callback) {
        _storageWriteArchiveCallbacks.push(callback);
    }
}

// storage_write_string migrated to new_storage.js

// delete strings in archive older than some amount of days
// clean up missing strings and zip text strings in archive while we're at it
function storage_clear_old_strings(days) {
	if (!(days >= 0)) return; // make sure it's valid before deleting everything by accident
	log("clearing strings older than " + days + " days");
	var ts_limit = Date.now() - 1000 * 60 * 60 * 24 * days;

	for (var i in archive) {
		var str = archive[i];
		
		log(str.ts, ts_limit);

		remove = (str.ts < ts_limit); // too old
		
		if (!remove && !fs.existsSync("data/string-" + str._id + ".zip")) { // zip doesn't exist
		
			if (fs.existsSync("data/string-" + str._id + ".txt")) { // text exists, try to make a zip from it
				try {
					log("zipping text string: " + str._id);
					var buf = fs.readFileSync("data/string-" + str._id + ".txt");
					var buffer = new Buffer(buf);
					var zip = zlibjs.deflateSync(buffer);
					fs.writeFileSync("data/string-" + str._id + ".zip", zip); // make the zip
					fs.unlinkSync("data/string-" + str._id + ".txt"); // remove the txt
				} catch (e) {
					console.log("error zipping text string: " + e);
				}
			}
			
			if (!fs.existsSync("data/string-" + str._id + ".zip")) { // zip still doesn't exist, just remove it
				log("unable to find/convert string: " + str._id);
				remove = true;
			}
		}
		
		// actually delete files or save the string
		if (remove) {
			if (fs.existsSync("data/string-" + str._id + ".txt")) fs.unlinkSync("data/string-" + str._id + ".txt");
			if (fs.existsSync("data/string-" + str._id + ".zip")) fs.unlinkSync("data/string-" + str._id + ".zip");
			delete archive[i];
		}

	}
	
	storage_write_archive(archive);
	
	log("done clearing old strings");

}

// remove symlink but do not crash if the file doesn't exist
function rm_symlinkSync(name) {
	try { fs.unlinkSync(name); } catch(e) {}; // ignore errors
//	fs.existsSync(name) && fs.unlinkSync(name); // this gets done if symlink target exists
//	fs.lstatSync(name) && fs.unlinkSync(name); // this works if symlink target does not exist
}

//function storage_import_targetfaces(file) { // migrated to new_storage:new_storage_import_targetfaces

function performanceLog(logmsg, start) {
var log = console.log;
    if (!_performanceLogEnabled) return 0;
    if (!start && !logmsg) return process.hrtime();
    if (logmsg && !start) {
        log('0ms - ' + logmsg);
        return process.hrtime();
    }
    var end = process.hrtime(start);
    var diff = Math.round((end[0]*1000) + (end[1]/1000000));
    if (logmsg) {
        log(diff + 'ms - ' + logmsg);
    }
    return diff;
}
